"""Reads files in s3, gs or azure buckets using fsspec and provides convenience resources for chunked reading of various file formats"""

from typing import Iterator, List, Optional, Tuple, Union

import dlt
from dlt.sources import DltResource
from dlt.sources.credentials import FileSystemCredentials
from dlt.sources.filesystem import FileItem, FileItemDict, fsspec_filesystem, glob_files

from .helpers import (
    AbstractFileSystem,
    FilesystemConfigurationResource,
)
from .readers import (
    ReadersSource,
    _read_csv,
    _read_csv_duckdb,
    _read_jsonl,
    _read_parquet,
)


@dlt.source(_impl_cls=ReadersSource, spec=FilesystemConfigurationResource)
def readers(
    bucket_url: str,
    credentials: Union[FileSystemCredentials, AbstractFileSystem],
    file_glob: Optional[str] = "*",
) -> Tuple[DltResource, ...]:
    """This source provides a few resources that are chunked file readers. Readers can be further parametrized before use
       read_csv(chunksize, **pandas_kwargs)
       read_jsonl(chunksize)
       read_parquet(chunksize)

    Args:
        bucket_url (str): The url to the bucket.
        credentials (FileSystemCredentials | AbstractFilesystem): The credentials to the filesystem of fsspec `AbstractFilesystem` instance.
        file_glob (str, optional): The filter to apply to the files in glob format. by default lists all files in bucket_url non-recursively
    """
    filesystem_resource = filesystem(bucket_url, credentials, file_glob=file_glob)
    filesystem_resource.apply_hints(
        incremental=dlt.sources.incremental("modification_date"),
    )
    return (
        filesystem_resource | dlt.transformer(name="read_csv")(_read_csv),
        filesystem_resource | dlt.transformer(name="read_jsonl")(_read_jsonl),
        filesystem_resource | dlt.transformer(name="read_parquet")(_read_parquet),
        filesystem_resource | dlt.transformer(name="read_csv_duckdb")(_read_csv_duckdb),
    )


@dlt.resource(
    primary_key="file_url", spec=FilesystemConfigurationResource, standalone=True
)
def filesystem(
    bucket_url: str = dlt.secrets.value,
    credentials: Union[FileSystemCredentials, AbstractFileSystem] = dlt.secrets.value,
    file_glob: Optional[str] = "*",
    files_per_page: int = 100,
    extract_content: bool = True,
) -> Iterator[List[FileItem]]:
    """This resource lists files in `bucket_url` using `file_glob` pattern. The files are yielded as FileItem which also
    provide methods to open and read file data. It should be combined with transformers that further process (ie. load files)

    Args:
        bucket_url (str): The url to the bucket.
        credentials (FileSystemCredentials | AbstractFilesystem): The credentials to the filesystem of fsspec `AbstractFilesystem` instance.
        file_glob (str, optional): The filter to apply to the files in glob format. by default lists all files in bucket_url non-recursively
        files_per_page (int, optional): The number of files to process at once, defaults to 100.
        extract_content (bool, optional): If true, the content of the file will be extracted if
            false it will return a fsspec file, defaults to False.

    Returns:
        Iterator[List[FileItem]]: The list of files.
    """

    if isinstance(credentials, AbstractFileSystem):
        fs_client = credentials
    else:
        fs_client = fsspec_filesystem(bucket_url, credentials)[0]

    files_chunk: List[FileItem] = []
    for file_model in glob_files(fs_client, bucket_url, file_glob):
        file_dict = FileItemDict(file_model, credentials)
        if extract_content:
            file_dict["file_content"] = file_dict.read_bytes()
        files_chunk.append(file_dict)  # type: ignore
        # wait for the chunk to be full
        if len(files_chunk) >= files_per_page:
            yield files_chunk
            files_chunk = []
    if files_chunk:
        yield files_chunk


read_csv = dlt.transformer(standalone=True)(_read_csv)
read_jsonl = dlt.transformer(standalone=True)(_read_jsonl)
read_parquet = dlt.transformer(standalone=True)(_read_parquet)
read_csv_duckdb = dlt.transformer(standalone=True)(_read_csv_duckdb)



from typing import List

from . import field

METRICS_SCHEMA = {
    "metrics.absolute_top_impression_percentage": "DOUBLE",
    "metrics.active_view_cpm": "DOUBLE",
    "metrics.active_view_ctr": "DOUBLE",
    "metrics.active_view_impressions": "INT64",
    "metrics.active_view_measurability": "DOUBLE",
    "metrics.active_view_measurable_cost_micros": "INT64",
    "metrics.active_view_measurable_impressions": "INT64",
    "metrics.active_view_viewability": "DOUBLE",
    "metrics.all_conversions": "DOUBLE",
    "metrics.all_conversions_by_conversion_date": "DOUBLE",
    "metrics.all_conversions_from_click_to_call": "DOUBLE",
    "metrics.all_conversions_from_directions": "DOUBLE",
    "metrics.all_conversions_from_interactions_rate": "DOUBLE",
    "metrics.all_conversions_from_interactions_value_per_interaction": "DOUBLE",
    "metrics.all_conversions_from_location_asset_click_to_call": "DOUBLE",
    "metrics.all_conversions_from_location_asset_directions": "DOUBLE",
    "metrics.all_conversions_from_location_asset_menu": "DOUBLE",
    "metrics.all_conversions_from_location_asset_order": "DOUBLE",
    "metrics.all_conversions_from_location_asset_other_engagement": "DOUBLE",
    "metrics.all_conversions_from_location_asset_store_visits": "DOUBLE",
    "metrics.all_conversions_from_location_asset_website": "DOUBLE",
    "metrics.all_conversions_from_menu": "DOUBLE",
    "metrics.all_conversions_from_order": "DOUBLE",
    "metrics.all_conversions_from_other_engagement": "DOUBLE",
    "metrics.all_conversions_from_store_visit": "DOUBLE",
    "metrics.all_conversions_from_store_website": "DOUBLE",
    "metrics.all_conversions_value": "DOUBLE",
    "metrics.all_conversions_value_by_conversion_date": "DOUBLE",
    "metrics.all_conversions_value_per_cost": "DOUBLE",
    "metrics.all_new_customer_lifetime_value": "DOUBLE",
    "metrics.asset_best_performance_cost_percentage": "DOUBLE",
    "metrics.asset_best_performance_impression_percentage": "DOUBLE",
    "metrics.asset_good_performance_cost_percentage": "DOUBLE",
    "metrics.asset_good_performance_impression_percentage": "DOUBLE",
    "metrics.asset_learning_performance_cost_percentage": "DOUBLE",
    "metrics.asset_learning_performance_impression_percentage": "DOUBLE",
    "metrics.asset_low_performance_cost_percentage": "DOUBLE",
    "metrics.asset_low_performance_impression_percentage": "DOUBLE",
    "metrics.asset_pinned_as_description_position_one_count": "INT64",
    "metrics.asset_pinned_as_description_position_two_count": "INT64",
    "metrics.asset_pinned_as_headline_position_one_count": "INT64",
    "metrics.asset_pinned_as_headline_position_three_count": "INT64",
    "metrics.asset_pinned_as_headline_position_two_count": "INT64",
    "metrics.asset_pinned_total_count": "INT64",
    "metrics.asset_unrated_performance_cost_percentage": "DOUBLE",
    "metrics.asset_unrated_performance_impression_percentage": "DOUBLE",
    "metrics.auction_insight_search_absolute_top_impression_percentage": "DOUBLE",
    "metrics.auction_insight_search_impression_share": "DOUBLE",
    "metrics.auction_insight_search_outranking_share": "DOUBLE",
    "metrics.auction_insight_search_overlap_rate": "DOUBLE",
    "metrics.auction_insight_search_position_above_rate": "DOUBLE",
    "metrics.auction_insight_search_top_impression_percentage": "DOUBLE",
    "metrics.average_cart_size": "DOUBLE",
    "metrics.average_cost": "DOUBLE",
    "metrics.average_cpc": "DOUBLE",
    "metrics.average_cpe": "DOUBLE",
    "metrics.average_cpm": "DOUBLE",
    "metrics.average_cpv": "DOUBLE",
    "metrics.average_impression_frequency_per_user": "DOUBLE",
    "metrics.average_order_value_micros": "INT64",
    "metrics.average_page_views": "DOUBLE",
    "metrics.average_target_cpa_micros": "INT64",
    "metrics.average_target_roas": "DOUBLE",
    "metrics.average_time_on_site": "DOUBLE",
    "metrics.benchmark_average_max_cpc": "DOUBLE",
    "metrics.benchmark_ctr": "DOUBLE",
    "metrics.biddable_app_install_conversions": "DOUBLE",
    "metrics.biddable_app_post_install_conversions": "DOUBLE",
    "metrics.bounce_rate": "DOUBLE",
    "metrics.clicks": "INT64",
    "metrics.combined_clicks": "INT64",
    "metrics.combined_clicks_per_query": "DOUBLE",
    "metrics.combined_queries": "INT64",
    "metrics.content_budget_lost_impression_share": "DOUBLE",
    "metrics.content_impression_share": "DOUBLE",
    "metrics.content_rank_lost_impression_share": "DOUBLE",
    "metrics.conversion_last_conversion_date": "DATE",
    "metrics.conversion_last_received_request_date_time": "DATE",
    "metrics.conversions": "DOUBLE",
    "metrics.conversions_by_conversion_date": "DOUBLE",
    "metrics.conversions_from_interactions_rate": "DOUBLE",
    "metrics.conversions_from_interactions_value_per_interaction": "DOUBLE",
    "metrics.conversions_value": "DOUBLE",
    "metrics.conversions_value_by_conversion_date": "DOUBLE",
    "metrics.conversions_value_per_cost": "DOUBLE",
    "metrics.cost_micros": "INT64",
    "metrics.cost_of_goods_sold_micros": "INT64",
    "metrics.cost_per_all_conversions": "DOUBLE",
    "metrics.cost_per_conversion": "DOUBLE",
    "metrics.cost_per_current_model_attributed_conversion": "DOUBLE",
    "metrics.cross_device_conversions": "DOUBLE",
    "metrics.cross_device_conversions_value_micros": "INT64",
    "metrics.cross_sell_cost_of_goods_sold_micros": "INT64",
    "metrics.cross_sell_gross_profit_micros": "INT64",
    "metrics.cross_sell_revenue_micros": "INT64",
    "metrics.cross_sell_units_sold": "DOUBLE",
    "metrics.ctr": "DOUBLE",
    "metrics.current_model_attributed_conversions": "DOUBLE",
    "metrics.current_model_attributed_conversions_from_interactions_rate": "DOUBLE",
    "metrics.current_model_attributed_conversions_from_interactions_value_per_interaction": "DOUBLE",
    "metrics.current_model_attributed_conversions_value": "DOUBLE",
    "metrics.current_model_attributed_conversions_value_per_cost": "DOUBLE",
    "metrics.eligible_impressions_from_location_asset_store_reach": "INT64",
    "metrics.engagement_rate": "DOUBLE",
    "metrics.engagements": "INT64",
    "metrics.general_invalid_click_rate": "DOUBLE",
    "metrics.general_invalid_clicks": "INT64",
    "metrics.gmail_forwards": "INT64",
    "metrics.gmail_saves": "INT64",
    "metrics.gmail_secondary_clicks": "INT64",
    "metrics.gross_profit_margin": "DOUBLE",
    "metrics.gross_profit_micros": "INT64",
    "metrics.historical_creative_quality_score": "ENUM",
    "metrics.historical_landing_page_quality_score": "ENUM",
    "metrics.historical_quality_score": "INT64",
    "metrics.historical_search_predicted_ctr": "ENUM",
    "metrics.hotel_average_lead_value_micros": "DOUBLE",
    "metrics.hotel_commission_rate_micros": "INT64",
    "metrics.hotel_eligible_impressions": "INT64",
    "metrics.hotel_expected_commission_cost": "DOUBLE",
    "metrics.hotel_price_difference_percentage": "DOUBLE",
    "metrics.impressions": "INT64",
    "metrics.impressions_from_store_reach": "INT64",
    "metrics.interaction_event_types": "ENUM",
    "metrics.interaction_rate": "DOUBLE",
    "metrics.interactions": "INT64",
    "metrics.invalid_click_rate": "DOUBLE",
    "metrics.invalid_clicks": "INT64",
    "metrics.lead_cost_of_goods_sold_micros": "INT64",
    "metrics.lead_gross_profit_micros": "INT64",
    "metrics.lead_revenue_micros": "INT64",
    "metrics.lead_units_sold": "DOUBLE",
    "metrics.linked_entities_count": "INT64",
    "metrics.linked_sample_entities": "STRING",
    "metrics.message_chat_rate": "DOUBLE",
    "metrics.message_chats": "INT64",
    "metrics.message_impressions": "INT64",
    "metrics.mobile_friendly_clicks_percentage": "DOUBLE",
    "metrics.new_customer_lifetime_value": "DOUBLE",
    "metrics.optimization_score_uplift": "DOUBLE",
    "metrics.optimization_score_url": "STRING",
    "metrics.orders": "DOUBLE",
    "metrics.organic_clicks": "INT64",
    "metrics.organic_clicks_per_query": "DOUBLE",
    "metrics.organic_impressions": "INT64",
    "metrics.organic_impressions_per_query": "DOUBLE",
    "metrics.organic_queries": "INT64",
    "metrics.percent_new_visitors": "DOUBLE",
    "metrics.phone_calls": "INT64",
    "metrics.phone_impressions": "INT64",
    "metrics.phone_through_rate": "DOUBLE",
    "metrics.publisher_organic_clicks": "INT64",
    "metrics.publisher_purchased_clicks": "INT64",
    "metrics.publisher_unknown_clicks": "INT64",
    "metrics.relative_ctr": "DOUBLE",
    "metrics.results_conversions_purchase": "DOUBLE",
    "metrics.revenue_micros": "INT64",
    "metrics.sample_best_performance_entities": "STRING",
    "metrics.sample_good_performance_entities": "STRING",
    "metrics.sample_learning_performance_entities": "STRING",
    "metrics.sample_low_performance_entities": "STRING",
    "metrics.sample_unrated_performance_entities": "STRING",
    "metrics.search_absolute_top_impression_share": "DOUBLE",
    "metrics.search_budget_lost_absolute_top_impression_share": "DOUBLE",
    "metrics.search_budget_lost_impression_share": "DOUBLE",
    "metrics.search_budget_lost_top_impression_share": "DOUBLE",
    "metrics.search_click_share": "DOUBLE",
    "metrics.search_exact_match_impression_share": "DOUBLE",
    "metrics.search_impression_share": "DOUBLE",
    "metrics.search_rank_lost_absolute_top_impression_share": "DOUBLE",
    "metrics.search_rank_lost_impression_share": "DOUBLE",
    "metrics.search_rank_lost_top_impression_share": "DOUBLE",
    "metrics.search_top_impression_share": "DOUBLE",
    "metrics.search_volume": "MESSAGE",
    "metrics.sk_ad_network_installs": "INT64",
    "metrics.sk_ad_network_total_conversions": "INT64",
    "metrics.speed_score": "INT64",
    "metrics.store_visits_last_click_model_attributed_conversions": "DOUBLE",
    "metrics.top_impression_percentage": "DOUBLE",
    "metrics.unique_users": "INT64",
    "metrics.units_sold": "DOUBLE",
    "metrics.valid_accelerated_mobile_pages_clicks_percentage": "DOUBLE",
    "metrics.value_per_all_conversions": "DOUBLE",
    "metrics.value_per_all_conversions_by_conversion_date": "DOUBLE",
    "metrics.value_per_conversion": "DOUBLE",
    "metrics.value_per_conversions_by_conversion_date": "DOUBLE",
    "metrics.value_per_current_model_attributed_conversion": "DOUBLE",
    "metrics.video_quartile_p100_rate": "DOUBLE",
    "metrics.video_quartile_p25_rate": "DOUBLE",
    "metrics.video_quartile_p50_rate": "DOUBLE",
    "metrics.video_quartile_p75_rate": "DOUBLE",
    "metrics.video_view_rate": "DOUBLE",
    "metrics.video_view_rate_in_feed": "DOUBLE",
    "metrics.video_view_rate_in_stream": "DOUBLE",
    "metrics.video_view_rate_shorts": "DOUBLE",
    "metrics.video_views": "INT64",
    "metrics.view_through_conversions": "INT64",
    "metrics.view_through_conversions_from_location_asset_click_to_call": "DOUBLE",
    "metrics.view_through_conversions_from_location_asset_directions": "DOUBLE",
    "metrics.view_through_conversions_from_location_asset_menu": "DOUBLE",
    "metrics.view_through_conversions_from_location_asset_order": "DOUBLE",
    "metrics.view_through_conversions_from_location_asset_other_engagement": "DOUBLE",
    "metrics.view_through_conversions_from_location_asset_store_visits": "DOUBLE",
    "metrics.view_through_conversions_from_location_asset_website": "DOUBLE",
}

METRIC_TO_DLT_TYPE = {
    "INT64": "bigint",
    "DOUBLE": "double",
    "STRING": "text",
    "ENUM": "text",
    # TODO: support message types
    # "MESSAGE": "string",
}


def dlt_metrics_schema(metrics: List[str]):
    """
    Returns a dictionary with only the metrics that are
    present in the given list of metrics.
    """
    schema = {}
    for metric in metrics:
        typ = METRICS_SCHEMA.get(metric)
        if typ is None:
            raise ValueError(f"Unsupported metric {metric}")

        if typ not in METRIC_TO_DLT_TYPE:
            raise ValueError(f"Unsupported metric '{metric}' of type '{typ}'")

        # ???: can we make these non-nullable?
        schema[field.to_column(metric)] = {
            "data_type": METRIC_TO_DLT_TYPE[typ],
        }
    return schema



"""Reads messages from Kinesis queue."""

from typing import Iterable, List, Optional

import dlt
from dlt.common import json, pendulum
from dlt.common.configuration.specs import AwsCredentials
from dlt.common.time import ensure_pendulum_datetime
from dlt.common.typing import StrStr, TAnyDateTime, TDataItem
from dlt.common.utils import digest128

from .helpers import get_shard_iterator, max_sequence_by_shard


@dlt.resource(
    name=lambda args: args["stream_name"],
    primary_key="kinesis_msg_id",
    standalone=True,
    max_table_nesting=0,
)
def kinesis_stream(
    stream_name: str,
    initial_at_timestamp: TAnyDateTime,
    credentials: AwsCredentials,
    last_msg: Optional[dlt.sources.incremental[StrStr]] = dlt.sources.incremental(
        "kinesis", last_value_func=max_sequence_by_shard
    ),
    max_number_of_messages: int = None,  # type: ignore
    milliseconds_behind_latest: int = 1000,
    parse_json: bool = True,
    chunk_size: int = 1000,
) -> Iterable[TDataItem]:
    """Reads a kinesis stream and yields messages. Supports incremental loading. Parses messages as json by default.

    Args:
        stream_name (str): The name of the stream to read from. If not provided, the
            value must be present in config/secrets
        credentials (AwsCredentials): The credentials to use to connect to kinesis. If not provided,
            the value from secrets or credentials present on the device will be used.
        last_msg (Optional[dlt.sources.incremental]): An incremental over a mapping from shard_id to message sequence
            that will be used to create shard iterators of type AFTER_SEQUENCE_NUMBER when loading incrementally.
        initial_at_timestamp (TAnyDateTime): An initial timestamp used to generate AT_TIMESTAMP or LATEST iterator when timestamp value is 0
        max_number_of_messages (int): Maximum number of messages to read in one run. Actual read may exceed that number by up to chunk_size. Defaults to None (no limit).
        milliseconds_behind_latest (int): The number of milliseconds behind the top of the shard to stop reading messages, defaults to 1000.
        parse_json (bool): If True, assumes that messages are json strings, parses them and returns instead of `data` (otherwise). Defaults to False.
        chunk_size (int): The number of records to fetch at once. Defaults to 1000.
    Yields:
            Iterable[TDataItem]: Messages. Contain Kinesis envelope in `kinesis` and bytes data in `data` (if `parse_json` disabled)

    """
    session = credentials._to_botocore_session()
    # the default timeouts are (60, 60) which is fine
    kinesis_client = session.create_client("kinesis")
    # normalize at_timestamp to pendulum
    initial_at_datetime = (
        None
        if initial_at_timestamp is None
        else ensure_pendulum_datetime(initial_at_timestamp)
    )
    # set it in state
    resource_state = dlt.current.resource_state()
    initial_at_datetime = resource_state.get(
        "initial_at_timestamp", initial_at_datetime
    )
    # so next time we request shards at AT_TIMESTAMP that is now
    resource_state["initial_at_timestamp"] = pendulum.now("UTC").subtract(seconds=1)

    shards_list = kinesis_client.list_shards(StreamName=stream_name)
    shards: List[StrStr] = shards_list["Shards"]
    while next_token := shards_list.get("NextToken"):
        shards_list = kinesis_client.list_shards(NextToken=next_token)
        shards.extend(shards_list)

    shard_ids = [shard["ShardId"] for shard in shards]

    # get next shard to fetch messages from
    while shard_id := shard_ids.pop(0) if shard_ids else None:
        shard_iterator, _ = get_shard_iterator(
            kinesis_client,
            stream_name,
            shard_id,
            last_msg,  # type: ignore
            initial_at_datetime,  # type: ignore
        )

        while shard_iterator:
            records = []
            records_response = kinesis_client.get_records(
                ShardIterator=shard_iterator,
                Limit=chunk_size,  # The size of data can be up to 1 MB, it must be controlled by the user
            )

            for record in records_response["Records"]:
                sequence_number = record["SequenceNumber"]
                content = record["Data"]

                arrival_time = record["ApproximateArrivalTimestamp"]
                arrival_timestamp = arrival_time.astimezone(pendulum.UTC)

                message = {
                    "kinesis": {
                        "shard_id": shard_id,
                        "seq_no": sequence_number,
                        "ts": ensure_pendulum_datetime(arrival_timestamp),
                        "partition": record["PartitionKey"],
                        "stream_name": stream_name,
                    },
                    "kinesis_msg_id": digest128(shard_id + sequence_number),
                }

                if parse_json:
                    message.update(json.loadb(content))
                else:
                    message["data"] = content
                records.append(message)
            yield records

            # do not load more  max_number_of_messages
            if max_number_of_messages is not None:
                max_number_of_messages -= len(records)
                if max_number_of_messages <= 0:
                    return

            # add child shards so we can request messages from them
            child_shards = records_response.get("ChildShards", None)
            if child_shards:
                for child_shard in child_shards:
                    child_shard_id = child_shard["ShardId"]
                    if child_shard_id not in shards:
                        shard_ids.append(child_shard_id)

            # gets 0 when no messages so we cutoff empty shards
            records_ms_behind_latest = records_response.get("MillisBehindLatest", 0)
            if records_ms_behind_latest < milliseconds_behind_latest:
                # stop taking messages from shard
                shard_iterator = None  # type: ignore
            else:
                # continue taking messages
                shard_iterator = records_response["NextShardIterator"]



from typing import Iterable

import dlt
from dlt.common.typing import TDataItem
from dlt.sources import DltResource

from ingestr.src.appsflyer.client import AppsflyerClient


@dlt.source(max_table_nesting=0)
def appsflyer_source(
    api_key: str, start_date: str, end_date: str
) -> Iterable[DltResource]:
    client = AppsflyerClient(api_key)

    @dlt.resource(write_disposition="merge", merge_key="install_time")
    def campaigns() -> Iterable[TDataItem]:
        yield from client.fetch_campaigns(start_date, end_date)

    @dlt.resource(write_disposition="merge", merge_key="install_time")
    def creatives() -> Iterable[TDataItem]:
        yield from client.fetch_creatives(start_date, end_date)

    return campaigns, creatives



from typing import Callable

from dlt.sources import DltResource, DltSource


def for_each(
    source: DltSource | DltResource, ex: Callable[[DltResource], None | DltResource]
):
    """
    Apply a function to each resource in a source.
    """
    if hasattr(source, "selected_resources") and source.selected_resources:
        resource_names = list(source.selected_resources.keys())
        for res in resource_names:
            ex(source.resources[res])  # type: ignore[union-attr]
    else:
        ex(source)  # type: ignore[arg-type]
